//*********************************************************************
// LCD generic display writer
// Author: Claudio Mignanti
// Web Site:
//
// Based on lcd function from: http://www.acmesystems.it/?id=15
//*********************************************************************

#include "stdio.h"     
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <asm/etraxgpio.h>
#include <stdarg.h>
#include <string.h>

#define HIGH IO_SETBITS
#define LOW IO_CLRBITS

int fd=0;
int pin_D4 = 8,
	pin_D5 = 9,
	pin_D6 = 10,
	pin_D7 = 11,
	pin_RS = 12,
	pin_E  = 13;
	

// Software delay in ms

void msDelay(int ms) {
  int i,a;
  int delayvar=10;
  
  for (a=0;a<ms;a++) {
    for (i=0;i<33084;i++) {
      delayvar*=2;        
      delayvar/=2;
    } 
  }
}   

// Software delay in us

void usDelay(int us) {
  int i,a;
  int delayvar=10;
  
  for (a=0;a<us;a++) {
    for (i=0;i<33;i++) {
      delayvar*=2;        
      delayvar/=2;
    } 
  }
}   

//*********************************************************************
// LCD functions
//*********************************************************************

// Set RS line
void lcd_rs(int status) {
  ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, status), 1 << pin_RS);
}


// Set E line
void lcd_e(int status) {
  ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, status), 1 << pin_E);
}

// Send an hi pulse to E line
void lcd_e_strobe() {
	lcd_e_hi(HIGH);
	lcd_e_lo(LOW);
}

// Send a nibble (4 bit) to LCD
void lcd_put_nibble(int value) {
	if (value&0x01) ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_SETBITS), 1 << pin_D4);
	else ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_CLRBITS), 1 << pin_D4);
	if (value&0x02) ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_SETBITS), 1 << pin_D5);
	else ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_CLRBITS), 1 << pin_D5);
	if (value&0x04) ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_SETBITS), 1 << pin_D6);
	else ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_CLRBITS), 1 << pin_D6);
	if (value&0x08) ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_SETBITS), 1 << pin_D7);
	else ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, IO_CLRBITS), 1 << pin_D7);
}

// Send a char to LCD
// data: Ascii char or instruction to send
// mode: 0 = Instruction, 1 = Data
void lcd_putc (unsigned char data, int mode) {
	int a;
	
	if (mode==0) lcd_rs(LOW);
	else lcd_rs(HIGH);
	
	a=(data>>4)&0x000F;
	lcd_put_nibble(a);
	lcd_e_strobe();
	a=data&0x000F;
	lcd_put_nibble(a);
	lcd_e_strobe();
} 

// Lcd initialization
void lcd_init() {
    unsigned long data, mask;

    data = mask = 0xFF << 8;
    ioctl(fd, _IO(ETRAXGPIO_IOCTYPE, 0x13), &mask);
    if ((mask & data) != data)  printf("Check your kernel config\n");

	lcd_rs(LOW);
	lcd_e(LOW);
	msDelay(15);
	
	lcd_put_nibble(0x03);
	lcd_e_strobe();
	msDelay(4);
	lcd_e_strobe();
	msDelay(2);
	lcd_e_strobe();
	msDelay(2);
	lcd_put_nibble(0x02);
	lcd_e_strobe();
	msDelay(1);

    lcd_putc(0x28,0);
	msDelay(1);
    lcd_putc(0x06,0);
	msDelay(1);
    lcd_putc(0x0C,0);
	msDelay(1);
    lcd_putc(0x01,0);
	msDelay(2);
} 

// Locate cursor on LCD
// row (0-2)
// col (1-39)
void lcd_locate(int row, int col) {
    lcd_putc(0x80+row*0x40+col,0);
    usDelay(35);
} 

// Clear LCD
void lcd_clear() {
  lcd_putc(0x01,0);
  msDelay(2);
} 

// Lcd version of printf
void lcd_printf(char *format, ...) {
  int i;
  
  va_list argptr;
  char buffer[1024];
  
  va_start (argptr, format);
  vsprintf (buffer, format, argptr);
  va_end (argptr);
  
  for (i=0; i < strlen(buffer); i++) {
    lcd_putc(buffer[i],1);
  }
}


/* lcd -f [conf_file] -x [xpos] -y [ypos] -t [text] -i -p [port] -c
 * where :
 * 	-i	inizialize the lcd
 *  -p is one of a, b, g (default g)
 *  -c clear lcd
 */
 
int main(int argc, char **argv) {
  	int x, y; 
  	char txt[255], port[255]="g", file[255]="/etc/lcd.conf";
  	char opt;
	while ((opt = getopt(argc, argv, "px:y:t:if:")) != -1) {
    	switch (opt) {
			case 'x':
            	x = atoi(optarg);
               	break;
            case 'y':
                y = atoi(optarg);
                break;
            case 't':
    	        strcpy(txt, optarg);
        	    break;
			case 'i':
                lcd_init();
                break;
			case 'p':
			   	strcpy(port, optarg);
			   	break;
	      case 'f':
    	        strcpy(file, optarg);
        	    break;
		}
	}
	
	char dev[20];
	scanf(dev, "/dev/gpio%c", port[0]); 
	
	if ((fd = open(dev, O_RDWR)) < 0) {
    	fprintf(stderr, "Error: Cannot open /dev/gpiog for read/write\n");
		return 1;
  	}
  
  	FILE *conf= fopen(file, O_RDONLY);
  	if (conf == 0) {
		fprintf(stderr, "Error: Cannot open %s read.\n", file);
		return 1;
	}
	
	char name[255]; 	//FIXME 
	int pin;
	
  	while (fscanf(conf, "%s %d\n",name , &pin ) != 0 ) {
		if (strcmp(name, "D4"))
			pin_D4 = pin;
		else if (strcmp(name, "D5"))
  			pin_D5 = pin;
		else if (strcmp(name, "D6"))
  			pin_D6 = pin;
		else if (strcmp(name, "D7"))
  			pin_D7 = pin;
		else if (strcmp(name, "RS"))
  			pin_RS = pin;
		else if (strcmp(name, "E"))
  			pin_E = pin;
  		else {
  			fprintf(stderr, "Error: in configuration file\n");
  			return 1;
  		}
	}

  	
  	if (!(pin_D4 || pin_D5 || pin_D6 || pin_D7 || pin_E || pin_RS)) {
		fprintf(stderr, "Error: not all pin configured!\n");
		return 1;  		
	} 
	
	lcd_locate(x,y);

    for (x=0; x < strlen(txt); x++) {
       lcd_putc(txt[x],1);
     }

	return 0;
}
